;; Problem 19
;; 14 June 2002

;; You are given the following information, but you may prefer to do
;; some research for yourself.

;;     * 1 Jan 1900 was a Monday.
;;     * Thirty days has September,
;;       April, June and November.
;;       All the rest have thirty-one,
;;       Saving February alone,
;;       Which has twenty-eight, rain or shine.
;;       And on leap years, twenty-nine.
;;     * A leap year occurs on any year evenly divisible by 4, but not
;;     on a century unless it is divisible by 400.

;; How many Sundays fell on the first of the month during the
;; twentieth century (1 Jan 1901 to 31 Dec 2000)?

;; day 0 is 1 jan 1900, a monday

(defun is-centry (year)
  (= 0 (mod year 100)))

(defun is-leap (year)
  (cond ((is-centry year)
	 (= 0 (mod year 400)))
	(t
	 (= 0 (mod year 4)))))

(loop for year from 1900 to 1910
      do
      (format t "year:~a : ~a~&" year (is-leap year)))

(assert (not (is-leap 1900)))
(assert (not (is-leap 1901)))
(assert (not (is-leap 1902)))
(assert (not (is-leap 1903)))
(assert      (is-leap 1904))
(assert      (is-leap 2000))
(assert (not (is-leap 2100)))

(defvar *days-per-month*
  '((january   31)
    (feburary  28)
    (march     31)
    (april     30)
    (may       31)
    (june      30)
    (july      31)
    (august    31)
    (september 30)
    (october   31)
    (november  30)
    (december  31)))

; (assert (= 365 (apply #'+ (mapcar #'second *days-per-month*))))

(defun next-day-of-week (day)
  (cond ((equalp 'monday    day) 'tuesday)
	((equalp 'tuesday   day) 'wednesday)
	((equalp 'wednesday day) 'thursday)
	((equalp 'thursday  day) 'friday)
	((equalp 'friday    day) 'saturday)
	((equalp 'saturday  day) 'sunday)
	((equalp 'sunday    day) 'monday)
	(t
	 (error "Unknown day: ~a~&" day))))

(defun day-number (day)
  (cond ((equalp 'monday    day) 1)
	((equalp 'tuesday   day) 2)
	((equalp 'wednesday day) 3)
	((equalp 'thursday  day) 4)
	((equalp 'friday    day) 5)
	((equalp 'saturday  day) 6)
	((equalp 'sunday    day) 7)
	(t
	 (error "Unknown day: ~a~&" day))))

; (next-day-of-week 'foo)

(defun next-mon (mon)
  (cond ((equalp 'january   mon) 'feburary)
	((equalp 'feburary  mon) 'march)
	((equalp 'march     mon) 'april)
	((equalp 'april     mon) 'may)
	((equalp 'may       mon) 'june)
	((equalp 'june      mon) 'july)
	((equalp 'july      mon) 'august)
	((equalp 'august    mon) 'september)
	((equalp 'september mon) 'october)
	((equalp 'october   mon) 'november)
	((equalp 'november  mon) 'december)
	((equalp 'december  mon) 'january)))

(defun month-number (mon)
  (cond ((equalp 'january   mon) 1)
	((equalp 'feburary  mon) 2)
	((equalp 'march     mon) 3)
	((equalp 'april     mon) 4)
	((equalp 'may       mon) 5)
	((equalp 'june      mon) 6)
	((equalp 'july      mon) 7)
	((equalp 'august    mon) 8)
	((equalp 'september mon) 9)
	((equalp 'october   mon) 10)
	((equalp 'november  mon) 11)
	((equalp 'december  mon) 12)
	(t
	 (error "Unknown month: ~a~&" mon))))

(defun days-per-month (mon year)
  (let ((pair (assoc mon *days-per-month*)))
    (unless pair
      (error "No such month: ~a / ~a~&" mon year))
    (if (and (equalp 'feburary mon)
	     (is-leap year))
	(+ 1 (second pair))
      (second pair))))

; (days-per-month 'feburary 1900)

;; How many Sundays fell on the first of the month during the
;; twentieth century (1 Jan 1901 to 31 Dec 2000)?

;; day 0 is 1 jan 1900, a monday

(defstruct (date)
  "Represents a simple date.  "
  (year 1900)
  (month 'january)
  (day 1)
  (day-of-month 1)
  (day-of-week 'monday)
  (day-of-year 1)
  (hours 0)
  (minutes 0)
  (seconds 0))

; (make-date)

(defun make-date-from-date (date)
  (make-date
   :year          (date-year date)
   :month 	  (date-month date)
   :day-of-month  (date-day-of-month date)
   :day-of-week   (date-day-of-week date)
   :day-of-year   (date-day-of-year date)
   :hours   	  (date-hours date)
   :minutes 	  (date-minutes date)
   :seconds 	  (date-seconds date)))

(defun new-date (&optional year month day-of-month day-of-week day-of-year hours minutes seconds)
  (make-date
   :year          year
   :month 	  month
   :day-of-month  day-of-month
   :day-of-week   day-of-week
   :day-of-year   day-of-year
   :hours   	  hours
   :minutes 	  minutes
   :seconds 	  seconds))

;(new-date 1900 'january 1 'monday 1 0 0 0)


(defun is-end-of-year (date)
  (and
   (equalp 'december (date-month date))
   (= 31 (date-day-of-month date))))

(assert (is-end-of-year (new-date 1900 'december 31 'monday 1 0 0 0)))

(defun next-date (date)
  (let ((new-date (make-date-from-date date))
	(end-of-year (is-end-of-year date)))
    ;(format t "new-date:~a~&" new-date)
    ;(format t "end-of-year:~a~&" end-of-year)
    (incf (date-day-of-month new-date))
    (incf (date-day-of-year new-date))
    (setf (date-day-of-week new-date)
	  (next-day-of-week (date-day-of-week new-date)))
    ;(format t "incremented dom, doy and dow~&")
    (when (> (date-day-of-month new-date)
	     (days-per-month (date-month new-date) (date-year new-date)))
      (setf (date-day-of-month new-date) 1)
      (setf (date-month new-date)
	    (next-mon (date-month new-date))))
    ;(format t "between whens...")
    (when end-of-year
	(incf (date-year new-date))
	(setf (date-day-of-year new-date) 1))
    new-date))

; (days-per-month (date-month (make-date)) 1900)

; (next-date (make-date))

(defun format-date (date)
  (format nil "~4,'0d~2,'0d~2,'0d ~3d ~a ~a"
	  (date-year date)
	  (month-number (date-month date))
	  (date-day-of-month date)
	  (date-day-of-year date)
	  (date-month date)
	  (date-day-of-week date)))

(defun format-datetime (date)
  (format nil "~4,'0d~2,'0d~2,'0d ~2,'0d:~2,'0d:~2,'0d"
	  (date-year date)
	  (month-number (date-month date))
	  (date-day-of-month date)
	  (date-hours date)
	  (date-minutes date)
	  (date-seconds date)))

;(format-date (make-date))
;(month-number (date-month (make-date)))
;(date-year (make-date))
;(date-day-of-month (make-date))
;(date-hours (make-date))

(defun write-date-file ()
  (with-open-file (out "/home/mortis/personal/projects/project-euler/dates.txt"
		       :direction :output 
		       :if-exists :supersede)
    (loop
     for date = (make-date) then (next-date date)
     do
     (when (<= 2001 (date-year date))
       (return t))
     (format out "~a~&" (format-date date)))))

; (time (write-date-file))

(defun get-jan-1901 ()
  (loop for d = (make-date) then (next-date d)
	do
	(when (< 1900 (date-year d))
	  (return d))))

; (format-date (get-jan-1901))

(time 
 (let ((num-first-sundays 0))
   (loop for date = (get-jan-1901) then (next-date date)
	 do
	 (when (<= 2001 (date-year date))
	   (return t))
	 (when (and (equalp 'sunday 
			    (date-day-of-week date))
		    (= 1 (date-day-of-month date)))
	   (incf num-first-sundays)))
   (format t "num-first-sundays:~a~&" num-first-sundays)
   num-first-sundays))
